from typing import Final

UNDEFINED: Final[str] = "undefined"

UTILIZATION_METRIC_INTERVAL: Final[float] = 5.0
UTILIZATION_METRIC_DETENTION: Final[float] = 600.0  # 10 minutes

CONTAINER_UTILIZATION_METRIC_LABEL_NAME: Final[str] = "container_metric_name"
DEVICE_UTILIZATION_METRIC_LABEL_NAME: Final[str] = "device_metric_name"
